;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
PWM=        0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone
;   el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y 
;   borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits
;   si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como
;   f_sck=f_clk/(2*(SPICTRL[7:0]+1))

;--------------------------------------
;----------- VECTORES -----------------
;-------------------------------------- 
		org 0			; RESET
		jr		start

		org		0x800
start:	
		ldi		r0,0		; 0.25 segundos de espera
		ldi		r1,46
del0:	subi	r0,r0,1
		jnz		del0
		subi	r1,r1,1
		jnz		del0

		ldpc	r6			; Reubicación
		word	endcode
		ldi		r7,0
		ldi		r2,endcode-inicode
reub1:	subi	r6,r6,1
		subi	r7,r7,1
		ld		r0,(r6)
		st		(r7),r0
		subi	r2,r2,1
		jnz		reub1

		ori		r0,r7,7		; Hexdump
		xori	r0,r0,7
		or		r1,r0,r0
		ldi		r2,((endcode-inicode)+7)/8
		adpc	r6,1
		jr		hexdump

		jind	r7

inicio:	
;----------------------------------------------

;----------------------------------------------
; Imprime R0 en hexadecimal
prthex:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r3
		;
		or		r2,r0,r0
		ldi		r3,4
prth1:	add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2	
		andi	r0,r2,15	; Los bits interesantes están en LSB
		cmpi	r0,10
		jnc		prth2
		addi	r0,r0,7
prth2:	ldi		r1,'0'
		add		r0,r0,r1
		adpc 	r6,1
		jr		putch
		subi	r3,r3,1
		jnz		prth1
		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;---------------------------------------------------
; Imprime R0 como 2 caracteres ASCII (little-endian)
; si es que son imprimibles. En caso de no serlo
; imprime '.'
putdchle:
		subi	r7,r7,1
		st		(r7),r6

		shr		r2,r0		; Ahorramos un desplazamiento
		ldi		r1,0xff
		and		r0,r0,r1
		ldi		r1,32
		cmp		r0,r1
		jc		pdch1
		ldi		r0,'.'
pdch1:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch2
		ldi		r0,'.'
pdch2:	adpc 	r6,1
		jr		putch
		shr		r0,r2
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		ldi		r1,32
		cmp		r0,r1
		jc		pdch3
		ldi		r0,'.'
pdch3:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch4
		ldi		r0,'.'
pdch4:	adpc 	r6,1
		jr		putch

		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;----------------------------------------------
; Hexadecimal dump
; R0: dir base
; R1: puntero a datos
; R2: Nº de filas de datos (8 datos por fila)

hexdump:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r5
		subi	r7,r7,1
		st		(r7),r4
		subi	r7,r7,1
		st		(r7),r3
		;
		or		r3,r0,r0	; r3: dir base
		or		r4,r1,r1	; r4: puntero a datos
		or		r5,r2,r2	; r5: contador de filas
hd1:	or		r0,r3,r3	; Imprimimos direccion
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '		; y dos espacios
		adpc 	r6,1
		jr		putch
		ldi		r0,' '
		adpc 	r6,1
		jr		putch

		subi	r7,r7,1		; Copia temporal de puntero
		st		(r7),r4
hd2:	ld		r0,(r4)		; Imprimimos los datos
		addi	r4,r4,1
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '
		adpc 	r6,1
		jr		putch
		andi	r0,r4,7
		jnz		hd2

		ldi		r0,' '		; y un espacio
		adpc 	r6,1
		jr		putch

		ld		r4,(r7)		; recuperamos el puntero
		addi	r7,r7,1		; para repetir los datos en ASCII little endian
hd3:	ld		r0,(r4)
		adpc 	r6,1
		jr		putdchle
		addi	r4,r4,1	
		andi	r0,r4,7
		jnz		hd3

		addi	r3,r3,8		; nueva línea
		ldi		r0,'\n'
		adpc 	r6,1
		jr		putch
		subi	r5,r5,1
		jnz		hd1

		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r4,(r7)
		addi	r7,r7,1
		ld		r5,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos

;------------------------------------------------------------
;--------- Código que se reubica al final de la RAM ---------
;------------------------------------------------------------
inicode:
; Lee un número hexadecimal
; Si comienza por @ es una dirección
gethex:	ldi		r2,0	; Dirección por defecto
gh0:	ldi		r4,0	; flag

gh1:	ldi		r3,4	; nº de digitos 

gh2:	ldi		r1,PFLAGS	; Lee caracter
gh3:	ld		r0,(r1)
		tsti	r0,1
		jz		gh3
		ldi		r1,UARTDAT
		ld		r0,(r1)
		st		(r1),r0	; eco

		ldi		r1,0x60	; Pasa a mayúsculas
		cmp		r0,r1
		jnc		gh10
		ldi		r1,0x20
		xor		r0,r0,r1

gh10:	ldi		r1,'0'
		cmp		r0,r1
		jnc		gh1		; si <'0' es retorno de linea, ignorar
		ldi		r1,'@'
		cmp		r0,r1	; Si '@' sigue una dirección para cargar(r4=flag)
		jnz		gh11
		or		r4,r1,r1
		jr		gh1
gh11:	ldi		r1,'X'
		cmp		r0,r1	; Si 'X' sigue una dirección para ejecutar (r4=flag)
		jnz		gh12
		or		r4,r1,r1
gh12:	ldi		r1,'G'  ; Las letras G-Z no son hexadecimales
		cmp		r0,r1
		jc		gh1

		; Dato hexadecimal
gh22:	ldi		r1,'A'
		cmp		r0,r1
		jnc		gh4
		subi	r0,r0,7
gh4:	ldi		r1,'0'
		sub		r0,r0,r1
		add		r5,r5,r5
		add		r5,r5,r5
		add		r5,r5,r5
		add		r5,r5,r5
		or		r5,r5,r0
		subi	r3,r3,1
		jnz		gh2

		ldi		r0,'@'
		cmp		r4,r0
		jnz		gh6
		or		r2,r5,r5	; Nueva dirección
		jr		gh0
gh6:	ldi		r0,'X'
		cmp		r4,r0
		jnz		gh7
		jind	r5			; Ejecuta desde dirección
gh7:	st		(r2),r5		; Dato: store
		addi	r2,r2,1		; 	  y dirección++
		jr		gh0

;-------- Final del código
endcode:

