;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
PWM=        0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone
;   el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y 
;   borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits
;   si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como
;   f_sck=f_clk/(2*(SPICTRL[7:0]+1))

;--------------------------------------
;----------- VECTORES -----------------
;--------------------------------------
		org 0			; RESET
		ldpc	r6
		word	inicio
		jind	r6

;		org 4			; IRQ0: externa
;		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
;		org	0xc			; IRQ2:	UART TX
;		jr		IRQ2
;		org	0x10		; IRQ3: UART RX
;		jr		IRQ3

		org		0x40
nintms:	word	0
nints:	word	0

IRQ1:	subi	r7,r7,1
		st		(r7),r0
irqe1:	subi	r7,r7,1
		st		(r7),r1

		ldi		r1,TIMER	; La lectura borra la IRQ del timer
		ld		r0,(r1)

		ldi		r1,nintms	; nintms++ 
		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0
		ldpc	r1			; if (nintms==1000)
		word	1000
		cmp		r0,r1
		jnz		lirq1
		ldi		r0,0
		ldi		r1,nintms	; nintms=0
		st		(r1),r0
		ldi		r1,nints	; nints++
		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0

		ldi		r1,GPOUT	; Contador Johnson en GPOUT
		ld		r0,(r1)
		add		r0,r0,r0
		tsti	r0,8
		jnz		lirq2
		ori		r0,r0,1		
lirq2:	st		(r1),r0

lirq1:	
		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
		reti
;-------------------------------------------------------------------
;		INICIO
;-------------------------------------------------------------------
;		org		0x1000

X0VAL=	-4080
Y0VAL=	2400
DELTA=  8
inicio:
		ldpc 	r7		; Puntero de Pila al final de RAM
		word	4095	; Una posición antes para la variable iterY
;		word	0xFFFF

		; LEDs en GPIO como salida
		ldi		r1,GPDIR
		ldi		r0,0x7
		st		(r1),r0
	
		ldi		r1,TIMER	; Interrupción cada ms
		ldpc	r0
		word	36000-1
		st		(r1),r0
		ld		r0,(r1)		; Borramos IRQ
		ldi		r1,IRQEN	; Habilita IRQ Timer
		ldi		r0,ETIMER
		st		(r1),r0

		ldi		r0,1		; Open graphics
		adpc 	r6,1		 
		jr		cmd
		ldi		r0,2		; cambio a texto
		adpc 	r6,1		 
		jr		cmd

restart:
		ldpc	r0			; Texto inicio 
		word	txt1
		adpc 	r6,1		
		jr		putsle	

		ldi		r0,3		; cambio a gráficos
		adpc 	r6,1		 
		jr		cmd

		ldi		r0,0		; reloj en 0
		ldi		r1,nintms
		st		(r1),r0
		ldi		r1,nints
		st		(r1),r0

		ldpc	r5			; y0 = 1
		word	Y0VAL
		ldpc	r0
		word	600
		st		(r7),r0		; iteraciones y

buc1:	ldpc	r4			; x0 = -2.34
		word	X0VAL	
		ldpc	r3			; iteraciones x		
		word	800
buc2:
		or		r0,r4,r4
		or		r1,r5,r5
		adpc	r6,1
		jr		mandel_iter
		adpc	r6,1			; El escape se manda por duplicado
		jr		putch_stuff

		ldi		r0,DELTA		; dx
		add		r4,r4,r0
		subi	r3,r3,1
		jnz		buc2

		ldi		r0,DELTA		; dy
		sub		r5,r5,r0
		ld		r0,(r7)
		subi	r0,r0,1
		st		(r7),r0
		jnz		buc1

mbuc:	
		ldi		r5,nints		; Tiempo en segundos
		ld		r5,(r5)

		ldi		r0,2			; cambio a texto
		adpc 	r6,1		 
		jr		cmd

		or		r0,r5,r5		; Imprime el tiempo
		adpc	r6,1
		jr		prtdec
		ldpc	r0
		word	txt2
		adpc	r6,1
		jr		putsle

		adpc	r6,1
		jr		getch

		ldi		r0,4		; CLS
		adpc 	r6,1		 
		jr		cmd

		jr		restart

txt1:	asczle "\nMandelbrot en gráficos remotos SDL\n"
txt2:	asczle " segundos\n"

;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos


;----------------------------------------------
; Recibe caracter del terminal
; parámetros:
;	R6: dirección de retorno
; retorna:
;	R0: dato recibido
; 	R1: modificado

getch:	ldi		r1,PFLAGS
getch1:	ld		r0,(r1)		; esperamos flag DV en 1
		tsti	r0,UARTDV
		jz		getch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro RX
		ld		r0,(r1)		; leemos dato
		jind	r6			; y retornamos

;---------------------------------------------------
; Envía dato con "stuffing"

putch_stuff:
		ldi		r1,0xFA
		cmp		r0,r1
		jnz		putch
		subi	r7,r7,1		; 0xFA se manda por duplicado
		st		(r7),r6
		adpc	r6,1
		jr		putch		; putch no cambia R0
		ld		r6,(r7)
		addi	r7,r7,1
		jr		putch

;---------------------------------------------------
; Envía comando gráfico
; R0: comando

cmd:	subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r0

		ldi		r0,0xFA
		adpc 	r6,1		
		jr		putch

		ld		r0,(r7)
		addi	r7,r7,1
		adpc 	r6,1		
		jr		putch

		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;----------------------------------------------
; Imprime R0 en decimal
; retorna: R0 y R1 modificados
; usa la pila para el almacenamiento temporal de dígitos

prtdec:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		;
		ldi		r1,0		; marca de final de cadena
		subi	r7,r7,1		; a la pila
		st		(r7),r1
prtd0:	ldi		r1,0		; R0=R0/10, R1=resto
		ldi		r6,16		; contador
prtd1:	add		r0,r0,r0
		adc		r1,r1,r1
		cmpi	r1,10
		jnc		prtd2
		subi	r1,r1,10
		addi	r0,r0,1
prtd2:	subi	r6,r6,1
		jnz		prtd1
		ldi		r6,'0'		; resto a la pila
		add		r1,r1,r6
		subi	r7,r7,1
		st		(r7),r1
		or		r0,r0,r0	; hasta que el cociente sea 0
		jnz		prtd0

prtd3:	ld		r0,(r7)		; caracter desde la pila
		jz		prtd9		; final de cadena?
		addi	r7,r7,1
		adpc	r6,1		; Imprime caracter
		jr		putch
		jr		prtd3

prtd9:	addi	r7,r7,1
		; epílogo
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;-------------------------------------------------------------------
;		Multiplicación fraccional
; Datos de 16 bits con 11 bits fraccionales val = dato/2048
;		R0 = (R0 * R1) >> 11
;  Nota: el valor de R2 se conserva
;-------------------------------------------------------------------

fmul11:	subi	r7,r7,1
		st		(r7),r6
		or		r6,r0,r0
		jpl		fml2			; Multiplicando (R6) siempre positivo
		neg		r6,r6
		neg		r1,r1
fml2:	ldi		r0,0
		add		r6,r6,r6		; Bit 14 (*8)
		jpl		fml3
		add		r0,r1,r1
		add		r0,r0,r0
		add		r0,r0,r0
fml3:	add		r6,r6,r6		; Bit 13 (*4)
		jpl		fml5
		add		r0,r0,r1
		add		r0,r0,r1
		add		r0,r0,r1
		add		r0,r0,r1
fml5:	add		r6,r6,r6		; Bit 12 (*2)
		jpl		fml7
		add		r0,r0,r1
		add		r0,r0,r1
		jr		fml7

fml4:	shra	r1,r1			; Resto de bits
fml7:	add		r6,r6,r6
		jz		fml6
		jpl		fml4
		add		r0,r0,r1
		jr		fml4
fml6:	ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;-------------------------------------------------------------------
; Iteración Mandelbrot
;  Calcula Zr = Zr*Zr - Zi*Zi + Cr
;          Zi = 2*Zr*Zi + Ci
;    mientras (Zr*Zr + Zi*Zi)<4
;  Parámetros: R0: Cr, R1: Ci
;  Retorna el nº de iteraciones (max 255)
;-------------------------------------------------------------------

mandel_iter:
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r5
		subi	r7,r7,1
		st		(r7),r4
		subi	r7,r7,1
		st		(r7),r3
		subi	r7,r7,4		; Variables i, zr2, zi2, zri en pila

		or		r4,r0,r0	; R4 = Cr
		or		r5,r1,r1	; R5 = Ci

		or		r2,r4,r4	; R2 = Zr = Cr
		or		r3,r5,r5	; R3 = Zi = Ci
		ldi		r0,0		; i = 0
		st		(r7),r0

mitr1:	or		r0,r2,r2	; zri=fmul11(zr,zi);
		or		r1,r3,r3
		adpc	r6,1
		jr		fmul11
		add		r0,r0,r0	; zri= 2*zr*zi
		addi	r1,r7,3
		st		(r1),r0

		or		r0,r2,r2	; zr2=fmul11(zr,zr);
		or		r1,r2,r2
		adpc	r6,1
		jr		fmul11
		addi	r1,r7,1
		st		(r1),r0

		or		r0,r3,r3	; zi2=fmul11(zi,zi);
		or		r1,r3,r3
		adpc	r6,1
		jr		fmul11
		addi	r1,r7,2
		st		(r1),r0

		addi	r1,r7,1		; escape: (zr2 + zi2) > (4<<11)
		ld		r1,(r1)
		add		r1,r1,r0
		ldpc	r6
		word	4<<11
		cmp		r1,r6
		jc		mitr9

		addi	r1,r7,1
		ld		r1,(r1)
		sub		r0,r1,r0
		add		r2,r0,r4	; Zr = Zr2 - Zi2 + Cr

		addi	r1,r7,3
		ld		r0,(r1)
		add		r3,r0,r5	; Zi = 2Zri + Ci

		ld		r0,(r7)		; i++
		addi	r0,r0,1
		st		(r7),r0
		ldi		r1,255
		cmp		r0,r1		; while (i<255)
		jnz		mitr1

mitr9:	ld		r0,(r7)		; return i
		addi	r7,r7,4
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r4,(r7)
		addi	r7,r7,1
		ld		r5,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6


