;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
PWM=        0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone
;   el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y 
;   borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits
;   si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como
;   f_sck=f_clk/(2*(SPICTRL[7:0]+1))

;--------------------------------------
;----------- VECTORES -----------------
;--------------------------------------
		org 0			; RESET
		jr		inicio

		org 4			; IRQ0: externa
		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
		org	0xc			; IRQ2:	UART TX
		jr		IRQ2
		org	0x10		; IRQ3: UART RX
		jr		IRQ3

		org		0x40
nirq:	word	0
siny1:	word	30000
siny2:	word	30000

;-------- IRQ TIMER -------

IRQ1:	subi	r7,r7,1		; registros modificados a la pila
		st		(r7),r0
		subi	r7,r7,1
		st		(r7),r1
		subi	r7,r7,1
		st		(r7),r2

		ldi		r1,nirq		; nirq++
		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0

timl1:	ldi		r1,siny1	; R2 = y1
		ld		r2,(r1)
		shra	r0,r2		; R0 = y1/32 (con signo)
		shra	r0,r0
		shra	r0,r0
		shra	r0,r0
		shra	r0,r0
		add		r2,r2,r2	; R0 = 2*y1 -y1/32
		sub		r2,r2,r0
		ldi		r1,siny2
		ld		r0,(r1)
		sub		r2,r2,r0	; R2 = y = 2*y1 -y1/32 -y2
		shra	r0,r2
		shra	r0,r0
		shra	r0,r0
		shra	r0,r0
		shra	r0,r0		; R0 = (y+32768)>>5 (de -1024 a +1023)
		ldi		r1,PWM
		st		(r1),r0		; nuevo valor PWM

		ldi		r1,siny1	; y2 = y1
		ld		r0,(r1)
		ldi		r1,siny2
		st		(r1),r0
		ldi		r1,siny1	; y1 = y
		st		(r1),r2

timl2:	ld		r2,(r7)		; recuperamos registros de la pila
		addi	r7,r7,1
		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
IRQ0:
IRQ2:
IRQ3:	reti

;-------------------------------------------------------------------
;		INICIO
;-------------------------------------------------------------------
inicio:
		ldpc 	r7			; Puntero de Pila al final de RAM
		word	4096

		ldpc	r0		; Imprime cadena de caracteres TXT
		word	txtle	; empaquetada tipo Little-Endian
		adpc r6,1		; Llamada a subrutina
		jr	putsle		

		ldi		r1,TIMER
		ldpc	r0
		word	1022
		st		(r1),r0
		ldi		r1,IRQEN
		ldi		r0,ETIMER
		st		(r1),r0

		ldi		r0,0x10
		ldi		r2,GPDIR
		st		(r2),r0

mbuc:	jr		.

txtle:	asczle "\nOnda senoidal de 991Hz en PWM\n"

;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6
;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos


;----------------------------------------------
; Recibe caracter del terminal
; parámetros:
;	R6: dirección de retorno
; retorna:
;	R0: dato recibido
; 	R1: modificado

getch:	ldi		r1,PFLAGS
getch1:	ld		r0,(r1)		; esperamos flag DV en 1
		tsti	r0,UARTDV
		jz		getch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro RX
		ld		r0,(r1)		; leemos dato
		jind	r6			; y retornamos


