;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX
SLEEP=		0x80	; A dormir si 1

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone
;   el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y 
;   borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits
;   si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como
;   f_sck=f_clk/(2*(SPICTRL[7:0]+1))

;--------------------------------------
;----------- VECTORES -----------------
;--------------------------------------
		org 0			; RESET
		ldpc	r6
		word	inicio
		jind	r6
		org 4			; IRQ0: externa
		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
		org	0xc			; IRQ2:	UART TX
		jr		IRQ2
		org	0x10		; IRQ3: UART RX
		jr		IRQ3

;--------------------------------------
;------------- CODIGO -----------------
;--------------------------------------
		org		0x40

inchar:	word	0
;--------------------------------------
; Interrupción de UART RX
;--------------------------------------

IRQ3:   subi    r7,r7,1 ; R0, R1 -> pila
        st      (r7),r0
        subi    r7,r7,1
        st      (r7),r1
        ldi     r1,UARTDAT
        ld      r0,(r1)
        ldi		r1,inchar
        st      (r1),r0
        ld      r1,(r7) ; R1, r0 <- pila
        addi    r7,r7,1
        ld      r0,(r7)
        addi    r7,r7,1
IRQ0:
IRQ1:
IRQ2:
        reti

;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6
;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos

;----------------------------------------------
;----------------------------------------------
;		MAIN
;----------------------------------------------
;----------------------------------------------
inicio:	ldi		r7,0

		ldi		r1,IRQEN	; Habilita UART RX
		ldi		r0,(EUARTDV)
		st		(r1),r0

mbuc:	ldpc	r0			; msg "pulsa"
		word	txt1
		adpc	r6,1
		jr		putsle

		ldi		r1,inchar	; Espera por IRQ RX
		ldi		r0,0
		st		(r1),r0
mb1:	ld		r0,(r1)
		jz		mb1

		ldpc	r0			; msg "ZZZ"
		word	txt11
		adpc	r6,1
		jr		putsle

		ldi		r1,IRQEN	
		ldi		r0,(EUARTDV|SLEEP)	; A dormir
		st		(r1),r0

		ldpc	r0			; msg "Desperté"
		word	txt2
		adpc	r6,1
		jr		putsle

		jr		mbuc

txt1:	asczle	"\nPulsa para dormir\n"
txt11:	asczle	"ZZZZ\n"
txt2:	asczle	"Desperté\n"


