;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como f_sck=f_clk/(2*(SPICTRL[7:0]+1))

		org 0			; RESET
;		jr	inicio
		ldpc	r6
		word	inicio
		jind	r6

		org 4			; IRQ0: externa
		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
		org	0xc			; IRQ2:	UART TX
		jr		IRQ2
		org	0x10		; IRQ3: UART RX
		jr		IRQ3

		org	64
IRQ1:	subi	r7,r7,1
		st		(r7),r0
		subi	r7,r7,1
		st		(r7),r1

		ldi		r1,TIMER	; La lectura borra la IRQ del timer
		ld		r0,(r1)

		ldpc	r1
		word	outreg

		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0
		ldi		r1,GPOUT
		adc		r0,r0,r0
		adc		r0,r0,r0
		adc		r0,r0,r0
		adc		r0,r0,r0
		adc		r0,r0,r0
		adc		r0,r0,r0
		st		(r1),r0

		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
IRQ0:
IRQ3:
		reti

IRQ2:	; UART THRE
		subi	r7,r7,1
		st		(r7),r0
		subi	r7,r7,1
		st		(r7),r1
		subi	r7,r7,1
		st		(r7),r2

		ldi		r2,UARTDAT
		ldpc	r1
		word	outdat
		ld		r0,(r1)
		jz		ir21
		st		(r2),r0
		ldi		r0,0
		st		(r1),r0
		jr		ir22

ir21:	ldi		r2,IRQEN
		ld		r0,(r2)
		andi	r0,r0,((~EUARTTHRE)&15)
		st		(r2),r0
ir22:
		ld		r2,(r7)
		addi	r7,r7,1
		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
		reti

;-------------------------------------------------------------------
;
;		INICIO
;
;-------------------------------------------------------------------
inicio:

		ldpc	r7			; Puntero de Pila al final de RAM
		word	4096

		; GPIO como salida
		ldi		r1,GPDIR
		ldi		r0,0x1F
		st		(r1),r0 

		; Prueba de interrupciones 

		ldi		r1,TIMER	; Interrupción cada 244us (4096 int/s)
		ldpc	r0
		word	64  ;8789-1
		st		(r1),r0
		ld		r0,(r1)		; Borramos IRQ
		ldi		r1,IRQEN	; Habilita IRQ Timer
		ldi		r0,ETIMER
		st		(r1),r0

		ldpc 	r0		; Imprime cadena de caracteres TXT
		word 	txtle	; empaquetada tipo Little-Endian
		adpc 	r6,1
		jr		putsle

		; volcado de la RAM interna
		ldi	r0,0
		ldi	r1,0
		ldi	r2,(buffer/8)
		adpc r6,1
		jr	hexdump

		ldpc	r0		; Imprime cadena de caracteres TXT
		word	txt2le	; empaquetada tipo Little-Endian
		adpc r6,1
		jr	putsle

		; Lectura de memoria flash
		ldi		r0,0
		ldpc	r1
		word	buffer
		ldpc	r2
		word	512
		adpc	r6,1
		jr		flashrd

		; Volcado de lo leído de la flash
		ldi		r0,0
		ldpc	r1
		word	buffer
		ldi		r2,64
		adpc 	r6,1
		jr		hexdump
		
buc:	adpc r6,1
		jr	getch
		cmpi r0,7		; crtl-G
		jz	0  			; Reinicio
		ldi	r1,'a'
		cmp r0,r1		; Carry en 0 si llevada al restar (r1>=r0 o r0 mayúscula)
		jnc	noch
		ldi	r1,0x20		; Cambiar bit 5
		xor	r0,r0,r1
noch:	adpc r6,1
		jr	putch
		jr	buc

;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6
;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch_orig:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos

putch:	ldpc	r1
		word	outdat
		ld		r1,(r1)		
		jnz		putch

		ldpc	r1
		word	outdat
		st		(r1),r0		; dato a cola tx

		ldi		r1,IRQEN
		ld		r0,(r1)
		ori		r0,r0,EUARTTHRE
		st		(r1),r0

		jind	r6			; y retornamos

;----------------------------------------------
; Recibe caracter del terminal
; parámetros:
;	R6: dirección de retorno
; retorna:
;	R0: dato recibido
; 	R1: modificado

getch:	ldi		r1,PFLAGS
getch1:	ld		r0,(r1)		; esperamos flag DV en 1
		tsti	r0,UARTDV
		jz		getch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro RX
		ld		r0,(r1)		; leemos dato
		jind	r6			; y retornamos


;----------------------------------------------
; Imprime R0 en hexadecimal
prthex:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r3
		;
		or		r2,r0,r0
		ldi		r3,4
prth1:	add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2	
		andi	r0,r2,15	; Los bits interesantes están en LSB
		cmpi	r0,10
		jnc		prth2
		addi	r0,r0,7
prth2:	ldi		r1,'0'
		add		r0,r0,r1
		adpc 	r6,1
		jr		putch
		subi	r3,r3,1
		jnz		prth1
		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;---------------------------------------------------
; Imprime R0 como 2 caracteres ASCII (little-endian)
; si es que son imprimibles. En caso de no serlo
; imprime '.'
putdchle:
		subi	r7,r7,1
		st		(r7),r6

		shr		r2,r0		; Ahorramos un desplazamiento
		ldi		r1,0xff
		and		r0,r0,r1
		ldi		r1,32
		cmp		r0,r1
		jc		pdch1
		ldi		r0,'.'
pdch1:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch2
		ldi		r0,'.'
pdch2:	adpc 	r6,1
		jr		putch
		shr		r0,r2
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		ldi		r1,32
		cmp		r0,r1
		jc		pdch3
		ldi		r0,'.'
pdch3:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch4
		ldi		r0,'.'
pdch4:	adpc 	r6,1
		jr		putch

		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;----------------------------------------------
; Hexadecimal dump
; R0: dir base
; R1: puntero a datos
; R2: Nº de filas de datos (8 datos por fila)

hexdump:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r5
		subi	r7,r7,1
		st		(r7),r4
		subi	r7,r7,1
		st		(r7),r3
		;
		or		r3,r0,r0	; r3: dir base
		or		r4,r1,r1	; r4: puntero a datos
		or		r5,r2,r2	; r5: contador de filas
hd1:	or		r0,r3,r3	; Imprimimos direccion
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '		; y dos espacios
		adpc 	r6,1
		jr		putch
		ldi		r0,' '
		adpc 	r6,1
		jr		putch

		subi	r7,r7,1		; Copia temporal de puntero
		st		(r7),r4
hd2:	ld		r0,(r4)		; Imprimimos los datos
		addi	r4,r4,1
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '
		adpc 	r6,1
		jr		putch
		andi	r0,r4,7
		jnz		hd2

		ldi		r0,' '		; y un espacio
		adpc 	r6,1
		jr		putch

		ld		r4,(r7)		; recuperamos el puntero
		addi	r7,r7,1		; para repetir los datos en ASCII little endian
hd3:	ld		r0,(r4)
		adpc 	r6,1
		jr		putdchle
		addi	r4,r4,1	
		andi	r0,r4,7
		jnz		hd3

		addi	r3,r3,8		; nueva línea
		ldi		r0,'\n'
		adpc 	r6,1
		jr		putch
		subi	r5,r5,1
		jnz		hd1

		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r4,(r7)
		addi	r7,r7,1
		ld		r5,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6


;-----------------------------------------------
;       SPI
;-----------------------------------------------
;	Transferencia SPI
;    dato a transmitir: R0
;    dato recibido: R0
;    modifica R1
SPIxfer:
		ldi		r1,SPIDAT
		st		(r1),r0
		ldi		r1,PFLAGS
bspx1:	ld		r0,(r1)
		tsti	r0,SPIBUSY
		jnz		bspx1
		ldi		r1,SPIDAT
		ld		r0,(r1)
		jind	r6			; retornamos	

;------------------------------------------------
; Lectura de Flash SPI
; R0: dir flash (64Kwords = 128 KBytes en Flash)
; R1: dir destino
; R2: nº de palabras

flashrd:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r4
		subi	r7,r7,1
		st		(r7),r3
		
		; SS1 en bajo, 16 bits, CLK/2
		ldi		r6,SPICTRL

		ldpc	r4			; Modo 16 bits, SS1
		word	0x8100		; Velocidad CLK/(2*(val+1)) = CLK/2
		st		(r6),r4

		or		r4,r1,r1	; R4=R1
		add		r3,r0,r0	; La flash espera direcciones de byte
		ldpc	r0			; R0=0x0300 (comado READ + MSBs de dir)
		word	0x0300
		adci	r0,r0,0		; Añadimos el acarreo de (dir*2)
		adpc 	r6,1		; SPIxfer({cmd,dir_msb})
		jr		SPIxfer
		or		r0,r3,r3	; R0=R3
		adpc 	r6,1		; SPIxfer(dir_lsb)
		jr		SPIxfer
flrd1:	adpc 	r6,1		; R0=SPIxfer(xxxx)
		jr		SPIxfer
		st		(r4),r0		; *R4++=R0
		addi	r4,r4,1
		subi	r2,r2,1
		jnz		flrd1

		;SS1 en alto
		ldi		r6,SPICTRL
		ldi		r0,0	
		st		(r6),r0
		
epi:		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r4,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

txtle:	asczle	"\n\n¡¡¡Hola Mundo!!!\n\nListado de la memoria interna\n"
txt2le:	asczle	"\n\nListado de la Flash SPI\n" 

outreg:	word 0x10
outdat:	word 0
		org	(.+7)&(~7)	; alineamos a múltiplo de 8
buffer:



